from .tokensets import TokenSet


_escape_pet_tokens = (
    ("{stop}", 0x03),
    ("{swlc}", 0x0E),
    ("{down}", 0x11),
    ("{rvon}", 0x12),
    ("{home}", 0x13),
    ("{del}", 0x14),
    ("{rght}", 0x1D),
    ("{sret}", 0x8D),
    ("{swuc}", 0x8E),
    ("{up}", 0x91),
    ("{rvof}", 0x92),
    ("{clr}", 0x93),
    ("{inst}", 0x94),
    ("{left}", 0x9D)
    )


class TokenSet_EscapePet(TokenSet):
    def __init__(self):
        super().__init__()
        self.raw = True
        self.add_tokens(_escape_pet_tokens)




class TokenTrie:
    def __init__(self):
        self.root = dict()

    def add_token(self, text, token):
        """Associate a value with a string token."""
        current = self.root
        for c in text:
            current = current.setdefault(c.upper(), {})
        current['_t_'] = token

    def next_token(self, text):
        """Return either text and value of the next token or a string which matches no token."""
        if text[0].upper() not in self.root:
            # initial letter not in trie, no match
            return text[0], None

        current = self.root
        match_stack = []
        for i in range(0, len(text)):
            if text[i].upper() not in current:
                if '_t_' in current:
                    # prefix in trie
                    return text[:i], current['_t_']
                if match_stack:
                    # there is a shorter match
                    return match_stack.pop()
                # no match
                return text[0], None
            if '_t_' in current:
                match_stack.append((text[:i], current['_t_']))
            current = current[text[i].upper()]
        return text, current.get('_t_')


class TokenSet:
    def __init__(self):
        self.token_to_string = dict()
        self.string_to_token = TokenTrie()
        self.raw = False
        self.skip_tokenize_after = ()
        self.secondary_token = None

    def add_tokens(self, tokens):
        # populate token trie
        for s, t in tokens:
            self.string_to_token.add_token(s, t)
        if self.raw:
            new_tokens = {t: s for s, t in tokens}
        else:
            new_tokens = {t: s.encode() for s, t in tokens}
        self.token_to_string.update(new_tokens)

    def set_secondary_tokens(self, lead_token, tokens):
        self.secondary_token = lead_token
        for s, t in tokens:
            self.string_to_token.add_token(s, bytes([lead_token, t]))
        self.secondary_token_to_string = {t: s.encode() for s, t in tokens}

    def tokenize(self, line, encoding):
        ret = bytearray()
        while line:
            match_text, token = self.string_to_token.next_token(line)
            line = line[len(match_text):]
            if token is None:
                # no token, just encode text
                ret += match_text.encode(encoding)
            else:
                if isinstance(token, int):
                    ret.append(token)
                    if token in self.skip_tokenize_after:
                        # remainder of line is not to be tokenized
                        return ret+line.encode(encoding)
                else:
                    ret += token

        return ret

    def expand(self, line_encoded, encoding):
        ret = ''
        previous = None
        for b in line_encoded:
            if b in self.token_to_string:
                if self.raw or not self.token_to_string[b].isascii():
                    ret += self.token_to_string[b]
                else:
                    ret += self.token_to_string[b].decode(encoding)
            elif self.secondary_token is not None and previous == self.secondary_token and b in self.secondary_token_to_string:
                ret += self.secondary_token_to_string[b].decode(encoding)
            elif b != self.secondary_token:
                ret += bytes([b]).decode(encoding)
            previous = b
        return ret


_token_set_register = {}


def lookup(set_name):
    """Return the token set for a given name."""
    if set_name in _token_set_register:
        return _token_set_register[set_name]
    raise LookupError("unknown token set: "+set_name)


def token_set_names():
    return _token_set_register.keys()



def vic20_expansion_required(start, length):
    """Return a tuple describing what memory expansion a memory range uses."""

    def _in_range(s, e, rs, re):
        return (e >= rs and e < re) or (s >= rs and s < re) or (s < rs and e > re)

    end = start + length

    x_3k = _in_range(start, end, 0x0400, 0x1000)
    x_blk1 = _in_range(start, end, 0x2000, 0x4000)
    x_blk2 = _in_range(start, end, 0x4000, 0x6000)
    x_blk3 = _in_range(start, end, 0x6000, 0x8000)
    x_blk5 = _in_range(start, end, 0xA000, 0xC000)

    return x_3k, x_blk1, x_blk2, x_blk3, x_blk5



import unittest

import cbm_files.expansion


class TestExpansion(unittest.TestCase):

    def test_expansion(self):
        self.assertEqual(cbm_files.expansion.vic20_expansion_required(0x1000, 0x250), (False, False, False, False, False))
        self.assertEqual(cbm_files.expansion.vic20_expansion_required(0x401, 0x1050), (True, False, False, False, False))
        self.assertEqual(cbm_files.expansion.vic20_expansion_required(0x1201, 0x2360), (False, True, False, False, False))
        self.assertEqual(cbm_files.expansion.vic20_expansion_required(0x1201, 0x4360), (False, True, True, False, False))
        self.assertEqual(cbm_files.expansion.vic20_expansion_required(0x1201, 0x6360), (False, True, True, True, False))
        self.assertEqual(cbm_files.expansion.vic20_expansion_required(0xA000, 0x1000), (False, False, False, False, True))



import codecs
import logging

log = logging.getLogger(__name__)


class DataFile:
    def __init__(self, fileh, encoding='petscii-c64en-uc'):
        self.lines = []
        try:
            _ = codecs.lookup(encoding)
        except LookupError:
            log.warning("PETSCII codecs not available, using ASCII")
            encoding = 'ascii'
        self.encoding = encoding

        line_encoded = bytearray()
        while fileh:
            b = fileh.read(1)
            if b == b'':
                # EOF
                break
            if b == b'\r':
                # CR
                self.lines.append(line_encoded)
                line_encoded = bytearray()
            else:
                line_encoded += b

        if line_encoded:
            line_encoded += b

    def from_text(self, fileh):
        """Create from Unicode text."""
        for line in fileh:
            line = line.rstrip('\n')
            try:
                self.lines.append(line.expandtabs().encode(self.encoding))
            except UnicodeEncodeError:
                raise ValueError("Unable to encode line: "+line)

    def to_binary(self):
        """Generator to return the file as binary."""
        for line in self.lines:
            yield line+b'\r'

    def to_text(self):
        """Generator to return lines as Unicode text."""
        for line in self.lines:
            yield line.decode(self.encoding)



from .basic_file import BASICFile  # noqa: F401
from .binary_file import BinaryFile  # noqa: F401
from .data_file import DataFile  # noqa: F401
from .pc64_file import PC64File  # noqa: F401
from .program_file import ProgramFile  # noqa: F401



from .tokensets import _token_set_register
from .tokenset_basic_v2 import TokenSet_BASICv2


_basic_ycxb_tokens = (
    ("MODE", 0xCC),
    ("CLB", 0xCD),
    ("CLS", 0xCE),
    ("CLG", 0xCF),
    ("AT", 0xD0),
    ("INK", 0xD1),
    ("AUX", 0xD2),
    ("VDU", 0xD3),
    ("VOL", 0xD4),
    ("CHAN", 0xD5),
    ("SOUND", 0xD6),
    ("SET", 0xD7),
    ("RSET", 0xD8),
    ("GCOL", 0xD9),
    ("HALT", 0xDA),
    ("UDG", 0xDB),
    ("CHAIN", 0xDC),
    ("PUT", 0xDD),
    ("RPT", 0xDE),
    ("INV", 0xDF),
    ("UPS", 0xE0),
    ("DNS", 0xE1),
    ("PLACE", 0xE2),
    ("CHAR", 0xE3),
    ("KEY", 0xE4)
    )


class TokenSet_BASIC_ycxb(TokenSet_BASICv2):
    """Tokens used by Your Computer Extended Basic."""
    def __init__(self):
        super().__init__()
        self.add_tokens(_basic_ycxb_tokens)


_token_set_register['basic-ycxb'] = TokenSet_BASIC_ycxb()



import unittest

from pathlib import Path
from unittest.mock import patch, mock_open, Mock

from cbm_files.pc64_file import PC64File


class TestMagic(unittest.TestCase):

    def test_good_magic(self):
        with patch.object(Path, 'open', mock_open(read_data=b'C64File\x00')):
            self.assertTrue(PC64File.is_valid_image(Path()))

    def test_bad_magic(self):
        with patch.object(Path, 'open', mock_open(read_data=b'INVALID')):
            self.assertFalse(PC64File.is_valid_image(Path()))


class TestPrg(unittest.TestCase):

    def test_prg(self):
        mock_file = Mock()
        mock_file.name = 'test.p00'
        mock_file.read.side_effect = [b'C64File\x00',
                                      b'TEST\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00']
        pc64 = PC64File(mock_file)
        self.assertEqual(pc64.name, b'TEST')
        self.assertEqual(pc64.file_type, 'PRG')


class TestRel(unittest.TestCase):

    def test_prg(self):
        mock_file = Mock()
        mock_file.name = 'test.r00'
        mock_file.read.side_effect = [b'C64File\x00',
                                      b'TEST\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x50']
        pc64 = PC64File(mock_file)
        self.assertEqual(pc64.name, b'TEST')
        self.assertEqual(pc64.file_type, 'REL')
        self.assertEqual(pc64.record_len, 80)


class TestSeq(unittest.TestCase):

    def test_prg(self):
        mock_file = Mock()
        mock_file.name = 'test.s01'
        mock_file.read.side_effect = [b'C64File\x00',
                                      b'TEST\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00']
        pc64 = PC64File(mock_file)
        self.assertEqual(pc64.name, b'TEST')
        self.assertEqual(pc64.file_type, 'SEQ')



import unittest

from cbm_files.tokenset_basic_v4 import TokenSet_BASICv4


class TestBasicV4(unittest.TestCase):
    def test_tokenize(self):
        tokenset = TokenSet_BASICv4()
        self.assertEqual(tokenset.tokenize('SYS', 'ascii'), b'\x9e')
        self.assertEqual(tokenset.tokenize('DIRECTORY', 'ascii'), b'\xda')

    def test_expand(self):
        tokenset = TokenSet_BASICv4()
        self.assertEqual(tokenset.expand(b'\x9e', 'ascii'), 'SYS')
        self.assertEqual(tokenset.expand(b'\xda', 'ascii'), 'DIRECTORY')



from .tokensets import _token_set_register
from .tokenset_basic_v2 import TokenSet_BASICv2


_vic20_super_expander_tokens = (
    ("KEY", 0xCC),
    ("GRAPHIC", 0xCD),
    ("SCNCLR", 0xCE),
    ("CIRCLE", 0xCF),
    ("DRAW", 0xD0),
    ("REGION", 0xD1),
    ("COLOR", 0xD2),
    ("POINT", 0xD3),
    ("SOUND", 0xD4),
    ("CHAR", 0xD5),
    ("PAINT", 0xD6),
    ("RPOT", 0xD7),
    ("RPEN", 0xD8),
    ("RSND", 0xD9),
    ("RCOLR", 0xDA),
    ("RGR", 0xDB),
    ("RJOY", 0xDC),
    ("RDOT", 0xDD)
    )


class TokenSet_VIC20_SuperExpander(TokenSet_BASICv2):
    """Tokens used by Super Expander (VIC-1211A)."""
    def __init__(self):
        super().__init__()
        self.add_tokens(_vic20_super_expander_tokens)


_token_set_register['vic20-super-expander'] = TokenSet_VIC20_SuperExpander()



from .tokensets import _token_set_register
from .tokenset_escape_pet import TokenSet_EscapePet


_escape_vic20_tokens = (
    ("{wht}", 0x05),
    ("{dish}", 0x08),
    ("{ensh}", 0x09),
    ("{red}", 0x1C),
    ("{grn}", 0x1E),
    ("{blu}", 0x1F),
    ("{f1}", 0x85),
    ("{f3}", 0x86),
    ("{f5}", 0x87),
    ("{f7}", 0x88),
    ("{f2}", 0x89),
    ("{f4}", 0x8A),
    ("{f6}", 0x8B),
    ("{f8}", 0x8C),
    ("{blk}", 0x90),
    ("{pur}", 0x9C),
    ("{yel}", 0x9E),
    ("{cyn}", 0x9F)
    )


class TokenSet_EscapeVIC20(TokenSet_EscapePet):
    def __init__(self):
        super().__init__()
        self.add_tokens(_escape_vic20_tokens)


_token_set_register['escape-vic20'] = TokenSet_EscapeVIC20()
_token_set_register['escape-vic1001'] = TokenSet_EscapeVIC20()



from .tokensets import _token_set_register
from .tokenset_basic_v2 import TokenSet_BASICv2


_basic_v4_tokens = (
    ("CONCAT", 0xCC),
    ("DOPEN", 0xCD),
    ("DCLOSE", 0xCE),
    ("RECORD", 0xCF),
    ("HEADER", 0xD0),
    ("COLLECT", 0xD1),
    ("BACKUP", 0xD2),
    ("COPY", 0xD3),
    ("APPEND", 0xD4),
    ("DSAVE", 0xD5),
    ("DLOAD", 0xD6),
    ("CATALOG", 0xD7),
    ("RENAME", 0xD8),
    ("SCRATCH", 0xD9),
    ("DIRECTORY", 0xDA),
    ("DCLEAR", 0xDB),
    ("BANK", 0xDC),
    ("BLOAD", 0xDD),
    ("BSAVE", 0xDE)
    )


class TokenSet_BASICv4(TokenSet_BASICv2):
    """Tokens used by BASIC 4.0."""
    def __init__(self):
        super().__init__()
        self.add_tokens(_basic_v4_tokens)


_token_set_register['basic-v4'] = TokenSet_BASICv4()



from setuptools import setup

setup(name='cbmfiles',
      description='Manipulate Commodore file types',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      use_scm_version=True,
      setup_requires=['setuptools_scm'],
      packages=['cbm_files'],
      author='Simon Rowe',
      author_email='srowe@mose.org.uk',
      url='https://eden.mose.org.uk/gitweb/?p=python-cbmfiles.git',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules'
        ]
      )



import codecs
import logging
import re
import struct

from . import tokensets
from . import tokenset_basic_v2  # noqa: F401
from . import tokenset_basic_v3_5  # noqa: F401
from . import tokenset_basic_v4  # noqa: F401
from . import tokenset_simons_basic  # noqa: F401
from . import tokenset_vic20_super_expander  # noqa: F401
from . import tokenset_vic20_super_expander_jp  # noqa: F401
from . import tokenset_escape_c64  # noqa: F401
from . import tokenset_escape_vic20  # noqa: F401


log = logging.getLogger(__name__)


class BASICFile:
    line_re = re.compile(r'(\d+) *(.*)')

    def __init__(self, fileh, start_addr=None, token_set='basic-v2', encoding='petscii-c64en-uc'):
        if start_addr is None and fileh is not None:
            # read start address from file
            self.start_addr, = struct.unpack('<H', fileh.read(2))
        else:
            self.start_addr = start_addr
        self.lines = {}

        self.token_set = token_set
        try:
            _ = codecs.lookup(encoding)
        except LookupError:
            log.warning("PETSCII codecs not available, using ASCII")
            encoding = 'ascii'
        self.encoding = encoding

        while fileh is not None:
            # link to start of next line (or NULL)
            link, = struct.unpack('<H', fileh.read(2))

            if link == 0:
                # final line, NUL follows
                break

            line_no, = struct.unpack('<H', fileh.read(2))

            # encoded line ends with NUL
            line_encoded = bytearray()
            while True:
                b = fileh.read(1)
                if b == b'\x00':
                    # EOL
                    break
                line_encoded += b

            self.add_encoded_line(line_no, line_encoded)

    def from_text(self, fileh):
        """Tokenize text file."""
        for line in fileh:
            m = self.line_re.match(line.strip())
            if m is None:
                raise ValueError("Syntax error in line: "+line)
            try:
                self.add_line(int(m.group(1)), m.group(2))
            except UnicodeEncodeError:
                raise ValueError("Unable to encode line: "+line)

    def to_binary(self, start_addr=None, prepend_addr=True):
        """Generator to return the program as binary."""
        if start_addr is None:
            start_addr = self.start_addr
        if prepend_addr:
            if start_addr is None:
                raise ValueError("No start address")
            yield struct.pack('<H', start_addr)

        for line_no, line_encoded in self.lines.items():
            start_addr += len(line_encoded) + 5
            yield struct.pack('<HH', start_addr, line_no) + line_encoded + b'\x00'

        yield b'\x00' * 2

    def to_text(self, text_ctrl=True):
        """Generator to return the program as Unicode text."""
        for line_no, line_encoded in self.lines.items():
            yield "{} {}".format(line_no, self.expand(line_encoded, text_ctrl))

    def merge(self, other):
        """Merge the lines of one program into another."""
        for line_no, line_encoded in other.lines.items():
            self.add_encoded_line(line_no, line_encoded)

    def add_line(self, line_no, line):
        """Add a line in the form of text."""
        self.add_encoded_line(line_no, self.tokenize(line))

    def delete_line(self, line_no):
        del self.lines[line_no]

    def add_encoded_line(self, line_no, line_encoded):
        """Add or replace a line in its correct position."""
        self.lines[line_no] = line_encoded
        self.lines = dict(sorted(self.lines.items(), key=lambda x: x[0]))

    @staticmethod
    def escape_set_name(encoding):
        """Return the name of the escape token set based on the encoding name."""
        if '-' in encoding:
            # format is 'petscii-<machine><lang>-<case>'
            machine = encoding.split('-')[1]
            return 'escape-'+machine[:-2]
        return 'escape-c64'

    def tokenize(self, line):
        """Convert a line of text to a binary encoded form."""
        tokenizer = tokensets.lookup(self.token_set)
        escaper = tokensets.lookup(self.escape_set_name(self.encoding))
        ret = bytearray()
        in_quote = False

        current = ''
        for c in line:
            if c == '"':
                if in_quote:
                    current += c
                    # encode text, always converting escapes
                    ret += escaper.tokenize(current, self.encoding)
                    current = ''
                else:
                    # tokenize text
                    ret += tokenizer.tokenize(current, self.encoding)
                    current = c

                in_quote = not in_quote
            else:
                current += c

        if current:
            if in_quote:
                ret += escaper.tokenize(current, self.encoding)
            else:
                ret += tokenizer.tokenize(current, self.encoding)

        return ret

    def expand(self, line_encoded, text_ctrl):
        """Convert a binary encoded line to text."""
        tokenizer = tokensets.lookup(self.token_set)
        escaper = tokensets.lookup(self.escape_set_name(self.encoding))
        ret = ''
        in_quote = False

        current = bytearray()
        for b in line_encoded:
            if b == ord('"'):
                if in_quote:
                    current.append(b)
                    if text_ctrl:
                        # expand escape sequences
                        ret += escaper.expand(current, self.encoding)
                    else:
                        ret += current.decode(self.encoding)
                    current = bytearray()
                else:
                    # expand tokens
                    ret += tokenizer.expand(current, self.encoding)
                    current = bytearray([b])

                in_quote = not in_quote
            else:
                current.append(b)

        if current:
            if in_quote:
                if text_ctrl:
                    ret += escaper.expand(current, self.encoding)
                else:
                    ret += current.decode(self.encoding)
            else:
                ret += tokenizer.expand(current, self.encoding)

        return ret



import os.path


class PC64File:
    PC64_MAGIC = b'C64File\x00'

    @staticmethod
    def is_valid_image(filepath):
        with filepath.open('rb') as fileh:
            magic = fileh.read(len(PC64File.PC64_MAGIC))

        return magic == PC64File.PC64_MAGIC

    def __init__(self, fileh):
        if fileh.read(len(self.PC64_MAGIC)) != self.PC64_MAGIC:
            raise ValueError("Invalid file format")

        header = fileh.read(0x12)
        if len(header) != 0x12:
            raise ValueError("Invalid file format")

        self.name = header[:0x10].rstrip(b'\x00')
        self.record_len = header[0x11]

        if self.record_len:
            self.file_type = 'REL'
        elif hasattr(fileh, 'name'):
            _, ext = os.path.splitext(fileh.name)
            type_map = {'P': 'PRG', 'S': 'SEQ', 'U': 'USR'}
            self.file_type = type_map.get(ext[1].upper())
        else:
            self.file_type = None
